#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
from ctypes import Structure, addressof, c_char_p, c_void_p
from ctypes.util import find_library

try:
    from ctypes import CDLL
except:
    pass

try:
    from ctypes import WinDLL
except:
    pass

lib = None

def r2lib():
    global lib
    if lib is not None:
        return lib
    lib_name = find_library("r_core")
    if lib_name is None:
        return None
    try:
        if sys.platform.startswith("win"):
            lib = WinDLL(lib_name)
        else:
            lib = CDLL(lib_name)
        return lib
    except OSError as err:
        # print(err)
        pass
    return None

class AddressHolder(object):
    def __get__(self, obj, type_):
        if getattr(obj, "_address", None) is None:
            obj._address = addressof(obj)
        return obj._address

    def __set__(self, obj, value):
        obj._address = value

class WrappedRMethod(object):
    def __init__(self, cname, args, ret):
        self.cname = cname
        self.args = args
        self.ret = ret
        self.args_set = False
        r2 = r2lib()
        if r2 is not None:
            self.method = getattr(r2, cname)
        else:
            raise ImportError("Cannot use ccall")

    def __call__(self, *a):
        if not self.args_set:
            if self.args:
                self.method.argtypes = [eval(x.strip()) for x in self.args.split(",")]
            self.method.restype = eval(self.ret) if self.ret else None
            self.args_set = True
        a = list(a)
        for i, argt in enumerate(self.method.argtypes):
            if argt is c_char_p:
                a[i] = a[i].encode()
        res = self.method(*a)
        if isinstance(res, bytes):
            return res.decode()
        return res


class WrappedApiMethod(object):
    def __init__(self, method, ret2, last):
        self.method = method
        self._o = None
        self.ret2 = ret2
        self.last = last

    def __call__(self, *a):
        result = self.method(self._o, *a)
        if self.ret2:
            if self.ret2 == "c_char_p":
                return result
            else:
                result = eval(self.ret2)(result)
        if self.last:
            return getattr(result, self.last)
        return result

    def __get__(self, obj, type_):
        self._o = obj._o
        return self


def register(cname, args, ret):
    ret2 = last = None
    if ret:
        if ret[0] >= "A" and ret[0] <= "Z":
            x = ret.find("<")
            if x != -1:
                ret = ret[0:x]
            last = "contents"
            ret = "POINTER(" + ret + ")"
        else:
            last = "value"
            ret2 = ret

    method = WrappedRMethod(cname, args, ret)
    wrapped_method = WrappedApiMethod(method, ret2, last)
    return wrapped_method, method

class RCore(Structure):  # 1
    def __init__(self):
        Structure.__init__(self)
        r2 = r2lib()
        if r2 is None:
            return
        r_core_new = r2.r_core_new
        r_core_new.restype = c_void_p
        self._o = r_core_new()

    _o = AddressHolder()

    cmd_str, r_core_cmd_str = register(
        "r_core_cmd_str", "c_void_p, c_char_p", "c_char_p"
    )
    free, r_core_free = register("r_core_free", "c_void_p", "c_void_p")


#  c = RCore()
#  c.cmd_str("o /bin/ls")
#  print(c)
#  print(c.cmd_str("s entry0;pd 20"))
#  c.free();



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""open_sync.py
This script use code from old __init__.py open object

"""

import re
import socket
import urllib
import os
from subprocess import Popen, PIPE
from r2pipe.open_base import OpenBase

def no_urlopen():
  raise IOError
try:
  from urllib.error import URLError
  from urllib.request import urlopen
except:
  URLError = IOError
  urlopen = no_urlopen


try:
    import fcntl
except ImportError:
    fcntl = None


class open(OpenBase):
    def __init__(self, filename="", flags=[], radare2home=None):
        super(open, self).__init__(filename, flags)
        self.pipe_read_sleep = 0.001
        self.pending = b''
        if filename.startswith("http://"):
            self._cmd = self._cmd_http
            self.uri = filename + "/cmd"
        elif filename.startswith("ccall://"):
            self._cmd = self._cmd_native
            self.uri = filename[8:]
        elif filename.startswith("tcp://"):
            r = re.match(r"tcp://(\d+\.\d+.\d+.\d+):(\d+)/?", filename)
            if not r:
                raise Exception("String doesn't match tcp format")
            self._cmd = self._cmd_tcp
            self.conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.conn.connect((r.group(1), int(r.group(2))))
        elif filename:
            self._cmd = self._cmd_process
            if radare2home is not None:
                if not os.path.isdir(radare2home):
                    raise Exception(
                        "`radare2home` passed is invalid, leave it None or put a valid path to r2 folder"
                    )
                r2e = os.path.join(radare2home, "radare2")
            else:
                r2e = "radare2"
            if os.name == "nt":
                # avoid errors on Windows when subprocess messes with name
                r2e += ".exe"
                hello_cmd = False
            else:
                hello_cmd = True
            cmd = [r2e, "-q0", filename]
            cmd = cmd[:1] + flags + cmd[1:]
            try:
                self.process = Popen(
                    cmd, shell=False, stdin=PIPE, stdout=PIPE, bufsize=0
                )
            except:
                raise Exception("ERROR: Cannot find radare2 in PATH")

            if hello_cmd:
                self.process.stdout.read(1)  # Reads initial \x00
                try:
                    self.process.stdin.write(("?V\n").encode("utf8"))
                    self.process.stdin.flush()
                    r = self.process.stdout
                    while True:
                        ch = r.read(1)
                        if ch == b'\x00':
                            break
                except:
                    raise Exception("ERROR: Cannot open %s" % filename)

    @staticmethod
    def __make_non_blocking(fd):
        if fcntl is not None:
            fl = fcntl.fcntl(fd, fcntl.F_GETFL)
            fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)
            return True

        if os.name != "nt":
            raise NotImplementedError()

        import msvcrt
        from ctypes import windll, byref
        from ctypes.wintypes import HANDLE, DWORD, BOOL

        try:
            from ctypes import POINTER
        except:
            from ctypes.wintypes import POINTER

        LPDWORD = POINTER(DWORD)
        SetNamedPipeHandleState = windll.kernel32.SetNamedPipeHandleState
        SetNamedPipeHandleState.argtypes = [HANDLE, LPDWORD, LPDWORD, LPDWORD]
        SetNamedPipeHandleState.restype = BOOL

        h = msvcrt.get_osfhandle(fd)

        PIPE_NOWAIT = DWORD(0x00000001)
        res = windll.kernel32.SetNamedPipeHandleState(h, byref(PIPE_NOWAIT), None, None)
        return res != 0

    def _cmd_process(self, cmd):
        cmd = cmd.strip().replace("\n", ";")
        try:
            self.process.stdin.write((cmd + "\n").encode("utf8"))
        except:
            return ''
        r = self.process.stdout
        self.process.stdin.flush()
        out = bytearray()
        foo = None
        while True:
            try:
                null_start = False
                if len(self.pending) > 0:
                    foo = self.pending
                    self.pending = b""
                else:
                    foo = r.read(4096)
                    if os.name == "nt":
                        if foo.startswith(b"\x00"):
                            foo = foo[1:]
                            null_start = True
                if foo:
                    zro = foo.find(b"\x00")
                    if zro != -1:
                        out += foo[0:zro]
                        if zro  < len(foo):
                            self.pending = foo[zro + 1:]
                        break
                    out += foo
                elif null_start:
                    break
            except:
                pass
        return out.decode("utf-8", errors="ignore")

    def _cmd_http(self, cmd):
        try:
            quocmd = urllib.parse.quote(cmd)
            response = urlopen("{uri}/{cmd}".format(uri=self.uri, cmd=quocmd))
            return response.read().decode("utf-8", errors="ignore")
        except URLError:
            pass
        return None

    def _cmd_tcp(self, cmd):
        res = b""
        self.conn.sendall(str.encode(cmd, "utf-8"))
        data = self.conn.recv(512)
        while data:
            res += data
            data = self.conn.recv(512)
        return res.decode("utf-8", errors="ignore")



# /usr/bin/env python3
# -*- coding: utf-8 -*-

"""open_base.py
base class for new open objects from open_sync and open_async. Code derived from __init__.py

"""

import os
import sys
import json
import shutil
import platform

from subprocess import Popen, PIPE

try:
    import r2lang
except ImportError:
    r2lang = None
try:
    from .native import RCore

    has_native = True
except ImportError:
    has_native = False

if os.name == "nt":
    from ctypes import byref, c_ulong, create_string_buffer, windll
    import msvcrt

    GENERIC_READ = 0x80000000
    GENERIC_WRITE = 0x40000000
    OPEN_EXISTING = 0x3
    INVALID_HANDLE_VALUE = -1
    PIPE_READMODE_MESSAGE = 0x2
    ERROR_PIPE_BUSY = 231
    ERROR_MORE_DATA = 234
    BUFSIZE = 4096
    chBuf = create_string_buffer(BUFSIZE)
    cbRead = c_ulong(0)
    cbWritten = c_ulong(0)


def in_rlang():
    return r2lang is not None and r2lang.cmd is not None


def jo2po(jo):
    from collections import namedtuple

    def _json_object_hook(d):
        return namedtuple("X", d.keys(), rename=True)(*d.values())

    def json2obj(data):
        return json.loads(data, object_hook=_json_object_hook)

    return json2obj(jo)


def get_radare_path():
    try:
        which = shutil.which
    except AttributeError:
        # In python 2 doesn't exist which function
        from distutils.spawn import find_executable

        which = find_executable

    bin_file = which("radare2")

    if bin_file and os.path.isfile(bin_file):
        return bin_file
    else:
        _platform = platform.system().lower()

        if _platform.startswith("darwin"):
            bin_file = "/usr/local/bin/radare2"
        else:
            bin_file = "/usr/bin/radare2"

        if os.path.isfile(bin_file):
            return bin_file
        else:
            raise IOError("radare2 can't be found in your system")


class OpenBase(object):
    """Class representing an r2pipe connection with a running radare2 instance
        Class body derived from __init__.py "open" class.
    
    
        """

    def __init__(self, filename="", flags=[]):
        """Open a new r2 pipe
                The 'filename' can be one of the following:

                * absolute or relative path to file
                * http://<host>:<port> to connect to an r2 webserver
                * tcp://<host>:<port> to connect to an r2 tcp server
                * #!pipe when launching it from r2 via RLang.pipe

                Args:
                    filename (str): path to filename or uri
                    flags (list of str): arguments, either in comapct form
                        ("-wdn") or sepparated by commas ("-w","-d","-n")
                Returns:
                    Returns an object with methods to interact with r2 via commands
                """
        self.use_cache = False
        self.cache = {}
        self.asyn = False
        if not filename and in_rlang():
            self._cmd = self._cmd_rlang
            return
        try:
            if os.name == "nt":
                mypipename = os.environ["r2pipe_path"]
                while 1:
                    hPipe = windll.kernel32.CreateFileW(
                        mypipename,
                        GENERIC_READ | GENERIC_WRITE,
                        0,
                        None,
                        OPEN_EXISTING,
                        0,
                        None,
                    )
                    if hPipe != INVALID_HANDLE_VALUE:
                        break
                    err = windll.kernel32.GetLastError()
                    print("Invalid Handle Value")
                    if err != ERROR_PIPE_BUSY:
                        print("Could not open pipe:", hex(err))
                        return
                    elif (windll.kernel32.WaitNamedPipeW(mypipename, 20000)) == 0:
                        print("Pipe busy")
                        return
                self.pipe = [hPipe, hPipe]
            else:
                self.pipe = [
                    int(os.environ["R2PIPE_IN"]),
                    int(os.environ["R2PIPE_OUT"]),
                ]
            self._cmd = self._cmd_pipe
            self.url = "#!pipe"
            return
        except:
            pass
        if filename.startswith("#!pipe"):
            raise Exception("ERROR: Cannot use #!pipe without R2PIPE_{IN|OUT} env")

    def invalidate_cache(self):
        self.cache = {}

    def _cmd_pipe(self, cmd):
        if self.use_cache and self.cache.get(cmd) is not None:
            return self.cache[cmd]
        out = b""
        cmd = cmd.strip().replace("\n", ";")
        if os.name == "nt":
            cmd = cmd.encode("utf-8")
            windll.kernel32.WriteFile(
                self.pipe[1], cmd, len(cmd), byref(cbWritten), None
            )
            while True:
                windll.kernel32.ReadFile(
                    self.pipe[1], chBuf, BUFSIZE, byref(cbRead), None
                )
                out += chBuf.value
                if ord(chBuf[cbRead.value - 1]) == 0:
                    if len(out) > 0 and out[-1] == 0:
                        out = out[0:-1]
                    break
        else:
            os.write(self.pipe[1], cmd.encode())
            while True:
                res = os.read(self.pipe[0], 4096)
                pos = res.find(b"\x00"[0])
                if pos != -1:
                    out += res[0:pos]
                else:
                    out += res
                if len(res) < 4096:
                    break
        res = out.decode("utf-8")
        if self.use_cache:
            self.cache[cmd] = res
        return res

    def _cmd_native(self, cmd):
        if self.use_cache and self.cache.get(cmd) is not None:
            return self.cache[cmd]
        cmd = cmd.strip().replace("\n", ";")
        if not has_native:
            raise Exception("No native ctypes connector available")
        if not hasattr(self, "native"):
            self.native = RCore()
            self.native.cmd_str("o " + self.uri)
        res = self.native.cmd_str(cmd)
        if isinstance(res, bytes):
            res = res.decode()
        if self.use_cache:
            self.cache[cmd] = res
        return res

    def _cmd_rlang(self, cmd):
        if self.use_cache and self.cache.get(cmd) is not None:
            return self.cache[cmd]
        res = r2lang.cmd(cmd)
        if self.use_cache:
            self.cache[cmd] = res
        return res

    def quit(self):
        """Quit current r2pipe session and kill
                """
        # HANGS self.cmd("q!!")
        if hasattr(self, "process"):
            import subprocess

            is_async = not isinstance(self.process, subprocess.Popen)
            if not is_async:
                for f in [self.process.stdin, self.process.stdout]:
                    if f is not None:
                        f.close()
            self.process.terminate()
            self.process.wait()
            delattr(self, "process")

            if is_async:
                import asyncio
                asyncio.get_event_loop().run_until_complete(asyncio.sleep(0.1))

    # r2 commands
    def cmd(self, cmd, **kwargs):
        """Run an r2 command return string with result
                Args:
                    cmd (str): r2 command
                Returns:
                    Returns an string with the results of the command
                
                res = self._cmd(cmd)
                if res is not None:
                    return res.strip()
                return None
                """

        res = self._cmd(cmd, **kwargs)
        if res is None:
            return None
        return res

    def cmdj(self, cmd, **kwargs):
        """Same as cmd() but evaluates JSONs and returns an object
                Args:
                    cmdj (str): r2 command
                Returns:
                    Returns a JSON object respresenting the parsed JSON
                """
        result = self.cmd(cmd, **kwargs)
        if result == "":
            result = "{}"
        try:
            data = json.loads(result)
        except (ValueError, KeyError, TypeError) as e:
            sys.stderr.write("r2pipe.cmdj.Error: %s\n" % (e))
            data = None
        return data

    def cmdJ(self, cmd, **kwargs):
        """Same as cmdj() but evaluates into a native Python Object
                Args:
                    cmdJ (str): r2 command
                Returns:
                    Returns a Python object respresenting the parsed JSON
                """
        result = self.cmd(cmd, **kwargs)
        try:
            return jo2po(result)
        except (ValueError, KeyError, TypeError) as e:
            sys.stderr.write("r2pipe.cmdj.Error: %s\n" % (e))
        return None

    def syscmd(self, cmd):
        """Executes a program and returns the output (stdout only)
                Args:
                    cmd (str): commandline shell command
                Returns:
                    Returns a string with the output
                """
        p = Popen(cmd, shell=True, stdin=PIPE, stdout=PIPE)
        out, err = p.communicate()
        return out

    def syscmdj(self, cmd):
        """Executes a program and returns an object representing the parsed JSON of the output
                Args:
                    cmd (str): commandline shell command
                Returns:
                    Returns an object constructed by parsing the JSON returned by the command
                """
        try:
            data = json.loads(self.syscmd(cmd))
        except (ValueError, KeyError, TypeError) as e:
            sys.stderr.write("r2pipe.syscmdj.Error %s\n" % (e))
            data = None
        return data
